/**
 * HardwareManager  07/14/2011
 *
 * This sends the Floomptagorpt samples to the sound hardware
 *
 * Copyright (C) 2011  Thomas G.Digby (bubbles@plergb.com)
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 *
 *   <http://www.gnu.org/licenses/>.
 *
 */

package floomptagorpt;

/**
 *
 * @author bubbles
 */

import javax.sound.sampled.*;


public class HardwareManager {

    SourceDataLine line;

    volatile Boolean wantToRun = false;  // start up waiting
    
    int hasRoomFor;
    int frameSize;

    int byteArraySize = 44100;  // size of byte array
    // must be a multiple of frame size = 4

    // line parameters
    //
    // standard CD format, with Intel endianness
    // probably supported on most systems
    float mySampleRate = 44100;
    int mySampleSizeInBits = 16;
    int myChannels = 2;
    boolean signedSamples = true;
    boolean bigLittleEndianSamples = false;

    Voices myVoices;
    PassBandManager myPassBandManager;


    // called at initial setup only
    String setupAudioLines(){

        String fate;
        AudioFormat format = new AudioFormat(mySampleRate, mySampleSizeInBits,
                myChannels, signedSamples, bigLittleEndianSamples);
        DataLine.Info info = new DataLine.Info(SourceDataLine.class, format);

        AudioFormat gotFormat;

        if (!AudioSystem.isLineSupported(info)){ // Handle the error.
            return null;
        }

        // Obtain and open the line.
        try {
            line = (SourceDataLine) AudioSystem.getLine(info);
            line.open(format);
        } catch (LineUnavailableException ex) {
   	    // Handle the error.
            return null;
        }

        int lineBufferSize = line.getBufferSize();
        hasRoomFor = line.available();

        gotFormat = line.getFormat();
        frameSize = gotFormat.getFrameSize();
        
        fate = "Format:  " + info.toString() + "\n" +
               "Line:  " + line.toString() + "\n\n" +

               "Line format " + gotFormat.toString() + "\n" +
               "Frame Size = " + frameSize + "\n\n" +

               "Line buffer size = " + lineBufferSize + "\n" +
               "Currently has room for " + hasRoomFor + " bytes \n\n" ;
        

        return fate;

    } // setupAudioLines()


    // called when user wants to start or resume playing the tone
    void setWantRun(){
        wantToRun = true;
    } // setWantRun()

    // called when user want to pause but not clear everything
    void setWantPause(){
        wantToRun = false;
    } // setWantPause()


    // cslled when stopping and wanting to start fresh
    void flushLineBuffer(){
        line.flush();
    } // flushLineBuffer

    int howManyArrays = 3;
    ArrayOfBytes[] byteArray;
    int writeThisArray;
    int howManyWritten;
    int byteOffset ;

    
    void setUpForSamples(int waveFormWanted,
            int howManyVoices, float freqRatio, double bottomFreq, boolean softInOut){

        byteArray = new ArrayOfBytes[howManyArrays];
        for(int i = 0; i < howManyArrays; i++)
            byteArray[i] = new ArrayOfBytes();

        writeThisArray = 0;
        howManyWritten = 0;
        byteOffset = 0;
        int howInOut = PassBandManager.FADE_NO;
        if(softInOut) howInOut = PassBandManager.FADE_YES;

        myVoices = new Voices();
        myVoices.setUpVoices(howManyVoices, freqRatio, bottomFreq);

        // choose waveform
        // default to sine 
        // may add custom later
        switch (waveFormWanted) {
            case (WaveFormSettings.WANT_SAWTH): myVoices.setUpSawtooth();
            break;
            case (WaveFormSettings.WANT_SQUARE): myVoices.setUpSquare();
            break;
            case (WaveFormSettings.WANT_TRIANG): myVoices.setUpTriangle();
            break;
            default: myVoices.setUpSine();
        } // switch

        myPassBandManager = new PassBandManager(
//                (int)bottomFreq, (int)myVoices.upperFreqLimit, 0);
                (int)myVoices.lowerFreqLimit, (int)myVoices.upperFreqLimit, 
                howInOut);
        myVoices.setPassBandManager(myPassBandManager);


    } // setUpForSamples()


    // called from loop in execution thread
    void generateNextSample() throws InterruptedException {

        int thisSample;
        byte thisByte;
        int bytesToWrite;

        if(wantToRun){ // deal with samples

            // fill all the byte arrays first
            // for each array that has room
            //
            // once the hardware is running
            // there should be no more than one array empty at any one time
            //
            // things should stay in order as long as sample generation
            // can keep up with the hardware

            for(int i = 0; i < howManyArrays; i++){
                if(!byteArray[i].dataLoaded){ // if not already loaded
                    // for each set of four bytes
                    for(int j = 0; j < byteArraySize; j += 4){
                        thisSample = myVoices.getNextSample();
                        // ls byte
                        thisByte = (byte) (thisSample & 0xFF);
                        byteArray[i].audioData[j] = thisByte; // both channels
                        byteArray[i].audioData[j + 2] = thisByte;
                        // ms byte
                        thisByte = (byte) ((thisSample >> 8)& 0xFF);
                        byteArray[i].audioData[j + 1] = thisByte; // both channels
                        byteArray[i].audioData[j + 3] = thisByte;
                    } // for each four bytes
                    byteArray[i].dataLoaded = true;
                } // if not already loaded
            } // for each array

            // if there are any full arrays, write them to hardware buffer
            bytesToWrite = line.available();

            if(bytesToWrite > 0){ // if room to write data

                // check for overflow
                if((bytesToWrite + byteOffset) > byteArraySize)
                    bytesToWrite = byteArraySize - byteOffset;

                // write the bytes
                howManyWritten =
                    line.write(byteArray[writeThisArray].audioData,
                    byteOffset, bytesToWrite);

                byteOffset += howManyWritten;
                // done with this array?
                if(byteOffset >= byteArraySize){

                    // ready for new data
                    byteArray[writeThisArray].dataLoaded = false;
                    // go to next array
                    writeThisArray++;
                    if(writeThisArray >= howManyArrays)
                        writeThisArray = 0;
                    byteOffset = 0;

                } // done with this array

            }  // if room to write data
            else Thread.sleep(10);
            // wait for a few hundred bytes so we don't hog the CPU

            // whether or not there is room to write, make sure
            // output hardware is running
            // if line isn't open, don't hog CPU
            if(line.isOpen()) line.start(); else Thread.sleep(100);

        } // want to run
        else { // don't want to run stop, wait a moment and try again
            if(line.isOpen()) line.stop();
            Thread.sleep(100);
        } // don't want to run

    } // generateNextSample


    class ArrayOfBytes {

        byte[] audioData = new byte[byteArraySize];
        // must be a multiple of frame size = 4

        boolean dataLoaded = false;

    } // class ArrayOfBytes


} // class HardwareManager

