/**
 * PassBandManager  07/14/2011
 *
 * Manages how Floomptagorpt voices come and go
 *  at the edges of the passband
 *
 * Copyright (C) 2011  Thomas G.Digby (bubbles@plergb.com)
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 *
 *   <http://www.gnu.org/licenses/>.
 *
 */

package floomptagorpt;

/**
 *
 * @author bubbles
 */

public class PassBandManager {

    // in case we add other fade-in-out options later
    public static final int PARABOLIC = 1;

    public static final int FADE_NO = 0;
    public static final int FADE_YES = PARABOLIC;

    // for scaling fractions values with ints
    public static final int SCALE_FACTOR = 30000;
    
    int minFreq, maxFreq;

    int[] passband;  // lookup table for attenuation factors


    public PassBandManager (int minFrequency, int maxFrequency, int howFade){

        int bufferPad = 100; // avoid array bounds errors
        
        minFreq =  minFrequency;
        maxFreq =  maxFrequency;

        double midLogScaleFreq;
        // double halfFreqBandRatio;
        double logHalfFreqBandRatio;
        double howFarFromCenter;
        double gainRatio;

        // one gain setting per frequency value
        passband = new int[maxFreq + bufferPad];

        // nothing below minimum freq
        for(int i = 0; i < minFreq; i++) passband[i] = 0;

        // apply function between min and max
        if(FADE_NO == howFade){ // no attenutation
            for(int i = minFreq; i < maxFreq; i++)
                    passband[i] = SCALE_FACTOR;
        } // if FADE_NO
        else if (PARABOLIC == howFade){ // parabolic-shaped function
            midLogScaleFreq = // geometric mean of limits
                    Math.pow(((double) minFreq * (double) maxFreq), 0.5);
            logHalfFreqBandRatio = Math.log(midLogScaleFreq / minFreq);
            for(int i = minFreq; i < maxFreq; i++){
                // parabola centered on log-scale center freq
                // gain = 1 at center, 0 at band edges
                howFarFromCenter = Math.log(midLogScaleFreq / i);
                gainRatio = 1 -
                        Math.pow(howFarFromCenter/logHalfFreqBandRatio, 2);
                passband[i] = (short) (gainRatio * SCALE_FACTOR);
            } // for i
        } // if PARABOLIC

        // nothing above max freq 
        for(int i = maxFreq; i < maxFreq + bufferPad; i++) passband[i] = 0;

    } // constructor

/*
    short getGain(int freq){ // get scaled attenuation for a given frquency

        return (short) passband[freq];

    } // getGain
*/

    int adjustAmplitude(int sample, int freq){

        int workingValue = sample * passband[freq];
        workingValue /= SCALE_FACTOR;

        return workingValue;
       
    } // adjustAmplitude()

    
} // class
