/**
 * Voices  07/14/2011
 *
 * Generates audio samples in Floomptagorpt application
 *
 * Copyright (C) 2011  Thomas G.Digby (bubbles@plergb.com)
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 *
 *   <http://www.gnu.org/licenses/>.
 *
 */

package floomptagorpt;

/**
 *
 * @author bubbles
 */

public class Voices {

    // eventually want a short value,
    // but start with int to make interim calculations easier

    // define one cycle of waveform, for grabbing samples
    static final int    ONE_CYCLE_ARRAY_LENGTH = 10000;
    int[]  oneCycle = new int[ONE_CYCLE_ARRAY_LENGTH];
    // close to but within bounds of short values
    static final int PEAK_AMPLITUDE_PLUS = 32767;
    static final int PEAK_AMPLITUDE_MINUS = -32767;
    static final int   PEAK_TO_PEAK_AMPLITUDE = 65534;

    float mySampleRate = 44100;  // assumes CD rate, should be able to override


    // 210 freq bumps rases freq by approx one octave
    // Any one bump should be too small to notice as a step
    // Start at one octave per second

    // maximum and minimum sweep rates in octaves per second
    static final double MAX_SWEEP_RATE = 10;
    static final double MIN_SWEEP_RATE = 0.0001;

    volatile int    bumpInterval = 210;         // bump every 210th sample initially
    volatile double freqBumpIncrement = 1.0033; // approx 210th root of 2

    double lowerFreqLimit;     // bottom of range
    double upperFreqLimit;     // top of range
    double freqLimitRatio;      // ratio of top of range to bottom
    double lowerStartFreq;     // 1/2 of one slot up from bottom
    
    // set in main setup
    int    howManyVoices;
    float voicesFreqRatio;

    // time to increment frequencies
    private int needToBump = 0;

    // output level setting
    // should be same as max on output Level slider
    public final static int OUT_LEVEL_DIVISOR = 100;
    volatile int outputLevelSetting = OUT_LEVEL_DIVISOR;

    oneVoice[] theTones; // individual frequency components

    PassBandManager myPassBandManager;


    void setUpVoices(int howMany, float ratio, double lowerLimit){
        // want a set of voices, each a given ratio above the last

        howManyVoices = howMany;
        voicesFreqRatio = ratio;
        lowerFreqLimit = lowerLimit;

        lowerStartFreq = lowerFreqLimit *
                Math.pow(voicesFreqRatio, 0.5);

        freqLimitRatio = Math.pow(voicesFreqRatio, howManyVoices);

        upperFreqLimit = lowerFreqLimit *
//                Math.pow(voicesFreqRatio, howManyVoices);
        freqLimitRatio;
        
        double myStartingFreq = lowerStartFreq;
        theTones = new oneVoice[howManyVoices];

        for(int i = 0; i < theTones.length; i++){
            theTones[i] = new oneVoice();
            theTones[i].setFrequency(myStartingFreq);
            myStartingFreq *= voicesFreqRatio;
        } // for

    } // setUpVoices


    void setPassBandManager(PassBandManager manager){
        myPassBandManager = manager;
    } // setPassBandManager


    // user will be allowed to choose one of these at run time

    // set up sample set for square wave
    void setUpSquare(){
        for(int i = 0; i < ONE_CYCLE_ARRAY_LENGTH/2; i++)
            oneCycle[i] = PEAK_AMPLITUDE_MINUS;
        for(int i = ONE_CYCLE_ARRAY_LENGTH/2; i < ONE_CYCLE_ARRAY_LENGTH; i++)
            oneCycle[i] =  PEAK_AMPLITUDE_PLUS;
    } // setUpSquare()

    // set up sample set for triangle wave
    void setUpTriangle(){
        for(int i = 0; i < ONE_CYCLE_ARRAY_LENGTH/2; i++)
            oneCycle[i] = (int) (PEAK_AMPLITUDE_MINUS + (
                    ((2.0f *
                    (float) PEAK_TO_PEAK_AMPLITUDE) / (float) ONE_CYCLE_ARRAY_LENGTH)
                    * i)
            );
        for(int i = ONE_CYCLE_ARRAY_LENGTH/2; i < ONE_CYCLE_ARRAY_LENGTH; i++)
            oneCycle[i] = (int) (PEAK_AMPLITUDE_PLUS - (
                    ((2.0f *
                    (float) PEAK_TO_PEAK_AMPLITUDE) / (float) ONE_CYCLE_ARRAY_LENGTH)
                    * (i - ONE_CYCLE_ARRAY_LENGTH/2))
            );
    } // setUpTriangle()

    // set up sample set for sawtooth wave
    void setUpSawtooth(){
        for(int i = 0; i < ONE_CYCLE_ARRAY_LENGTH; i++)
            oneCycle[i] = (int) (PEAK_AMPLITUDE_MINUS + (
                    ((float) PEAK_TO_PEAK_AMPLITUDE / (float) ONE_CYCLE_ARRAY_LENGTH)
                    * i)
            );
    } // setUpSawtooth()

    // set up sample set for sine wave
    void setUpSine(){
        for(int i = 0; i < ONE_CYCLE_ARRAY_LENGTH; i++)
            oneCycle[i] = (int) (
                (double) PEAK_AMPLITUDE_PLUS * Math.sin(
                2 * Math.PI * ((double) i / (double) ONE_CYCLE_ARRAY_LENGTH)
            ));
     } // setUpSine()


    // called from execution thread
    short getNextSample(){

        int theSample = 0;

        // time to bump the frequency?
        needToBump++;
        if(needToBump >= bumpInterval){
            needToBump = 0;
            for(oneVoice item : theTones){
                item.bumpFrequency();
            }
        } // need to bump

        // get the actual sample
        for(oneVoice item : theTones){
            theSample += item.getSample();
        }  // get the actual sample

        // don't let values add to overload
//        return (short) (theSample / howManyVoices);
        // also adjust out0ut level
        return (short) ((theSample * outputLevelSetting) /
             (howManyVoices * OUT_LEVEL_DIVISOR));

    } // getNextSample


    void setSweepRate(double octavesPerSecond){

        double bumpsPerSecond = 200;  // default rate for 1 octave/second

        if(octavesPerSecond > MAX_SWEEP_RATE) // out of range
            octavesPerSecond = MAX_SWEEP_RATE;
        if(octavesPerSecond < -MAX_SWEEP_RATE)
            octavesPerSecond = -MAX_SWEEP_RATE;

        if((octavesPerSecond < MIN_SWEEP_RATE)&&
            (octavesPerSecond > -MIN_SWEEP_RATE)){ // treat as zero
                freqBumpIncrement = 1.0;  // no change to frequency
                bumpInterval = 10000;     // don't waste time with it
        } // treat as zero
        else {  // valid value

            bumpsPerSecond =
                (bumpsPerSecond * Math.pow(Math.abs(octavesPerSecond), 0.5));

            bumpInterval = (int) (mySampleRate / bumpsPerSecond);

            freqBumpIncrement = Math.pow(
                (Math.pow(2, octavesPerSecond)), (1 / bumpsPerSecond));

        } // valid value

    } // setSweepRate

    
    void setOutputLevel(int level){
        outputLevelSetting = level;
    } // setOutputLevel


    // generates one tone of the set
    class oneVoice {

        // use double for stride to avoid quantizing too soon
        double myFrequency;
        double myStride;
        double myPointer = 0;

        
        // get one sample value, to be mixed with other voices, etc.
        int getSample(){

            // get the sample
            int mySample = oneCycle[(int) myPointer];

            // bump pointer for next time, wrapping if necessary
            myPointer += myStride;
            while (myPointer >= ONE_CYCLE_ARRAY_LENGTH)
                myPointer -= ONE_CYCLE_ARRAY_LENGTH;
            if(myPointer < 0) myPointer = 0;

            // adjust amplitude via frequency window

//            return mySample;

            return myPassBandManager.adjustAmplitude(
                    mySample, (int) myFrequency);

        } // getSample()


        // to raise or lower frequency every so often
        void bumpFrequency(){

            // multiply by bump increment
            // then wrap around
            // if below lower limit, wrap to  upper
            // if above upper limit, wrap to loweer

            myFrequency *= freqBumpIncrement;
            if(myFrequency < lowerFreqLimit) {
                myFrequency *= freqLimitRatio;}
            if(myFrequency > upperFreqLimit) {
                myFrequency /= freqLimitRatio;}

            myStride = ((myFrequency / mySampleRate) * ONE_CYCLE_ARRAY_LENGTH);

        } // bumpFrequency()


        // to set the initial frequency
        void setFrequency(double frequency){

            myFrequency = frequency;
            myStride = ((myFrequency / mySampleRate) * ONE_CYCLE_ARRAY_LENGTH);

        } // setFrequency()

    } // class oneVoice

} //  class Voices
